import asyncio
import os, signal
import subprocess
import time
from typing import List

from common_util import StrEnum


display_number = 99
display_number_id = f":{display_number}"

custom_env = os.environ.copy()
custom_env["DISPLAY"] = str(display_number_id)


class ProcessName(StrEnum):
    """Processes."""

    XVFB = "XVFB"
    DESKTOP = "DESKTOP"
    GLINT = "GLINT"


class NamedPopen(subprocess.Popen):
     def __init__(self, *args, name=None, owner=None, **kwargs):
         self.name = name
         self.owner = owner
         super().__init__(*args, **kwargs)


class GlintSessionHandler:
    def __init__(self):
        self.processes: List[NamedPopen] = []

    async def start_process_monitoring(self):
        while True:
            dead_processes = [proc for proc in self.processes if proc.poll() is not None]
            live_processes = [proc for proc in self.processes if proc.poll() is None]
            self.processes = live_processes

            for proc in dead_processes:
                print(f"[DEBUG]>>> {proc.args}  {proc.pid} terminated")  # noqa: T201
                self._on_process_finished(proc)

            await asyncio.sleep(0.1)

    def _on_process_finished(self, proc):
        # Если граф окружение завершает работу, это может значит, что юзер нажал на кноку (в ОС) Завершить сессию.
        # В это случае завершаем все процессы
        if proc.name == ProcessName.DESKTOP:
            self.terminate_all_processes(wait_for_finish=False, owner=proc.owner)

    def init_session(self, username):
        self.terminate_all_processes()

        # let user use sound card
        subprocess.run(["usermod", "-aG", "audio", username])
        subprocess.run(["mkhomedir_helper", username])


    def check_proc(self, process):
        if process.errors or not process.pid:
            print(f"[DEBUG]>>> {process.errors}")  # noqa: T201
            process.terminate()
            return False
        return True

    def start_proc(self, cmd, env=None, append=True, name=None, owner=None):
        if name:
            print(f"[DEBUG]>>> Starting {name}")  # noqa: T201
        try:
            proc = NamedPopen(cmd, env=env, preexec_fn=os.setsid, name=name, owner=owner)
            result = self.check_proc(proc)
            if result and append:
                self.processes.append(proc)
            return result
        except Exception as err:  # noqa: PIE786
            print(f"[DEBUG]>>> {err}")  # noqa: T201
            return False

    def start_xvfb(self, username, width, height, depth):
        try:
            os.remove(f"/tmp/.X11-unix/X{display_number}")
            os.remove(f"/tmp/.X{display_number}-lock")
        except FileNotFoundError:
            pass

        cmd = f"Xvfb {display_number_id} -screen 0 {width}x{height}x{depth}"
        is_running = self.start_proc(["sudo", "-u", username] + cmd.split(), name=ProcessName.XVFB, owner=username)
        if not is_running:
            is_running = self.start_proc(["su", "-", username, "-c", cmd], name=ProcessName.XVFB, owner=username)

        time.sleep(1)
        return is_running

    def start_desktop(self, username, desktop):
        run_data = ["sudo", "-u", username, desktop]
        is_running = self.start_proc(run_data, custom_env, name=ProcessName.DESKTOP, owner=username)

        time.sleep(0.5)
        return is_running

    def start_glint(self, username):
        cmd = ["sudo", "-u", username, "bash", "/opt/glint-server/start_glint.sh"]
        is_running = self.start_proc(cmd, custom_env, name=ProcessName.GLINT, owner=username)

        time.sleep(0.5)
        return is_running

    def start_pulseaudio(self, username):
        cmd = ["sudo", "-u", username, "pulseaudio"]
        is_running = self.start_proc(cmd, custom_env)
        time.sleep(0.5)
        return is_running

    def get_process(self, name, owner=None):
        try:
            if owner:
                return next(proc for proc in self.processes if proc.name == name and proc.owner == owner)
            else:
                return next(proc for proc in self.processes if proc.name == name)
        except StopIteration:
            return None

    def terminate_process(self, name, wait_for_finish=True, owner=None):

        proc = self.get_process(name, owner)
        if proc is None:
            return

        # Soft
        try:
            os.kill(os.getpgid(proc.pid), signal.SIGTERM)
            if wait_for_finish:
                proc.wait(0.3)  # Very important to wait untill process finish before starting new session
        except ProcessLookupError:
            print(f"[DEBUG]>>> Failed to stop process {proc.args}  {proc.pid}")
        except subprocess.TimeoutExpired:
            print(f"[DEBUG]>>> Failed to stop process {proc.args}  {proc.pid}. Timeout expired")
        # Hard
        #try:
        #    os.kill(os.getpgid(proc.pid), signal.SIGKILL)
        #except ProcessLookupError:
        #    pass

    def terminate_all_processes(self, wait_for_finish=True, owner=None):
        print("[DEBUG]>>> terminate_all_processes", wait_for_finish, owner)
        # Делаем так потому что для корректного завершения сначала нужно завершить графические процессы,
        # а потом уже сам граф сервер
        self.terminate_process(ProcessName.GLINT, wait_for_finish, owner)
        self.terminate_process(ProcessName.DESKTOP, wait_for_finish, owner)
        self.terminate_process(ProcessName.XVFB, wait_for_finish, owner)


glint_session_handler = GlintSessionHandler()

