import asyncio
import os

import uvicorn
from cysystemd import daemon
from cysystemd.daemon import Notification
from fastapi import FastAPI
from fastapi.middleware.httpsredirect import HTTPSRedirectMiddleware
from uvicorn.main import Server

from auth_handlers import auth_router
from common_util import cancel_async_task
from domain_socket_listener import (
    MsgFromGlintServer,
    create_subscriber,
    get_socket_msg,
    listen_for_unix_sock_connections,
    stop_unix_sock_listening,
)
from general_info_handlers import general_info_router
from glint_session import glint_session_handler
from settings import settings

app = FastAPI()
app.include_router(auth_router)
app.include_router(general_info_router)
app.add_middleware(HTTPSRedirectMiddleware)


original_exit_handler = Server.handle_exit


def handle_exit_code(*args, **kwargs):
    print("Stopping service", flush=True)  # noqa: T201
    glint_session_handler.terminate_all_processes()
    settings.is_running = False
    stop_unix_sock_listening()
    original_exit_handler(*args, **kwargs)
    print("Service stopped", flush=True)  # noqa: T201
    os._exit(0)


Server.handle_exit = handle_exit_code


async def handle_unix_sock_messages():
    domain_socket_msg_queue = create_subscriber()

    while settings.is_running:
        msg = await get_socket_msg(domain_socket_msg_queue)
        if msg == MsgFromGlintServer.USER_CONNECTED.value:
            pass
        elif msg == MsgFromGlintServer.USER_DISCONNECTED.value:
            glint_session_handler.terminate_all_processes()


@app.on_event("startup")
async def service_tasks_startup():
    print("Starting up ...")  # noqa: T201

    loop = asyncio.get_event_loop()
    asyncio.ensure_future(loop.run_in_executor(None, listen_for_unix_sock_connections))
    loop.create_task(handle_unix_sock_messages())

    # Process state monitoring
    loop.create_task(glint_session_handler.start_process_monitoring())

    # Tell systemd that our service is ready
    try:
        daemon.notify(Notification.READY, return_exceptions=False)
    except RuntimeError:
        print("Failed to notify daemon")
    print("Startup complete", flush=True)  # noqa: T201


if __name__ == "__main__":

    # Start server
    uvicorn.run(
        "main:app",
        host="0.0.0.0",
        port=7778,
        log_level="info",
        ssl_keyfile="cert/Private.key",
        ssl_certfile="cert/Certificate.crt",
    )

# Launch
# PYTHONPATH=.
# launch: python3 main.py
